#!/usr/bin/env python
"""Normalizes the functional MRI data based on the noise level and concatenates
results.

Based on the variance_normalize.m script from Saad Jbabdi
"""
import numpy as np
from nibabel import cifti2
from scipy import linalg

from mcot.cifti import cmd


def run(data, demean=True):
    """Normalizes the functional MRI data.

    :param data: (..., N) array of fMRI data with N timepoints
    :param demean: if True demean the data as well as normalizing
    :return: (..., N) array of normalized and potentially demeaned data

    Example matlab code from variance_normalize.m by Saad Jbabdi
    function yn = variance_normalise(y)
    % yn = variance_normalise(y)
    % y is TxN

    addpath ~steve/matlab/groupPCA/
    yn=y;
    [uu,ss,vv]=ss_svds(y,30);
    vv(abs(vv)<2.3*std(vv(:)))=0;
    stddevs=max(std(yn-uu*ss*vv'),0.001);
    yn=yn./repmat(stddevs,size(yn,1),1);
    """
    res = data.copy()
    if demean:
        res -= res.mean(-1)[..., None]
    u, s, v = linalg.svd(res.reshape(-1, res.shape[-1]), full_matrices=False)
    # threshold spatial maps
    u[abs(u) < 2.3 * np.std(u, 0)] = 0
    sdiag = np.diag(s, 0)
    residuals = res - u.dot(sdiag.dot(v))
    res /= np.std(residuals)
    return res.reshape(data.shape)


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    res = []
    ref_axes = None
    for arr, axes in args.input:
        res.append(run(arr.T).T)
        if ref_axes is None:
            ref_axes = axes
        else:
            assert ref_axes == axes
    full_arr = np.concatenate(res, 0)
    series = ref_axes[0]
    new_series = cifti2.SeriesAxis(series.start, series.step, full_arr.shape[0], series.unit)
    args.output((full_arr, (new_series, ref_axes[1])))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument('output', type=cmd.output, help='normalized output dataset')
    parser.add_argument('input', type=cmd.greyordinate_in, nargs='+', help='one or more functional MRI dataset')



#!/usr/bin/env python
"""Correlates two CIFTI files."""
import os.path as op
import time

import numpy as np
from loguru import logger
from nibabel import cifti2
from scipy import stats


def correlate_arr(arr1, arr2, collapse=False):
    """Computes the Pearson r correlation for two arrays.

    :param arr1: (M, N) array for N observations of M variables
    :param arr2: (M, K) array for K obsevations of M variables
    :param collapse: assume N==K and only return (N, ) array with the r correlation
    :return: (N, K) array with the r correlation (or (N, ) array if collapse is set)
    """
    arr1 = np.atleast_2d(arr1)
    arr2 = np.atleast_2d(arr2)
    if collapse:
        assert arr1.shape == arr2.shape
    demeaned1 = arr1 - arr1.mean(0)
    demeaned2 = arr2 - arr2.mean(0)
    if collapse:
        r_num = np.einsum('ij,ij->j', demeaned1, demeaned2)
        r_den = np.sqrt(np.sum(demeaned1 ** 2, 0) * np.sum(demeaned2 ** 2, 0))
    else:
        r_num = demeaned1.T.dot(demeaned2)
        r_den = np.sqrt(
                np.sum(demeaned1 ** 2, 0)[:, None] * np.sum(demeaned2 ** 2, 0)[None, :]
        )
    return r_num / r_den


class FakeDConn(object):
    """Pretends to be a nibabel image of the dconn file."""
    def __init__(self, dtseries):
        self.dataobj = FakeDConnDataObject(dtseries)


class FakeDConnDataObject(object):
    """Pretends to be a nibabel dataobject of a dconn file."""
    def __init__(self, dtseries):
        self.dtseries = dtseries

    def __getitem__(self, item):
        slc1, slc2 = item
        return correlate_arr(
                self.dtseries.dataobj[:, slc1],
                self.dtseries.dataobj[:, slc2]
        )


def correlate(dconn1, dconn2, select_features=slice(None)):
    """Correlates two memory mapped arrays.

    :param dconn1: (M, N) nibabel Cifti2Image (M features for N greyordinates)
    :param dconn2: (M, N) nibabel Cifti2Image (M features for N greyordinates)
    :param select_features: which features to select
    :return: (N, ) in-memory array
    """
    logger.info('Correlating {}'.format(select_features))
    ngrey = dconn1.shape[1]
    step = int(np.ceil(1e8 / ngrey))
    res = np.zeros(ngrey)
    loading_time = 0
    correlate_time = 0
    for idx in range(0, ngrey, step):
        logger.debug('Calculating correlation for grayordinates between {} and {}'.format(idx, idx + step))
        start_time = time.time()
        if isinstance(select_features, list) or isinstance(select_features, tuple):
            local1 = np.concatenate([dconn1.dataobj[sf, idx:idx + step] for sf in select_features], 0)
            local2 = np.concatenate([dconn2.dataobj[sf, idx:idx + step] for sf in select_features], 0)
        else:
            local1 = np.array(dconn1.dataobj[select_features, idx:idx + step])
            local2 = np.array(dconn2.dataobj[select_features, idx:idx + step])
        mid_time = time.time()
        loading_time += mid_time - start_time
        for sub_idx in range(local1.shape[1]):
            res[idx + sub_idx] = stats.spearmanr(local1[:, sub_idx], local2[:, sub_idx])[0]
        correlate_time += time.time() - mid_time
    logger.info('Loading data took {} minutes'.format(loading_time / 60))
    logger.info('Correlating the data took {} minutes'.format(correlate_time / 60))
    return res


def run(dconn1, dconn2, split_greyordinates: cifti2.BrainModelAxis=None):
    """Computes the correlation between dense connectomes.

    :param dconn1: (M, N) nibabel Cifti2Image
    :param dconn2: (M, N) nibabel Cifti2Image
    :param split_greyordinates: defines M voxels/vertices to select sub-set of features the features
    :return: dictionary mapping names of target regions to (N, ) arrays
    """
    logger.info('Computing full correlation')

    res = {'full': correlate(dconn1, dconn2)}
    if split_greyordinates is None:
        return res

    cortex = []
    sub_cortical = []
    left = []
    right = []
    for struc_name, struc_slice, struc_bm in split_greyordinates.iter_structures():
        short_name = struc_name.split('_', maxsplit=2)[2]
        if 'CORTEX' in short_name:
            logger.info('Computing correlation relative to %s', short_name)
            res[short_name] = correlate(dconn1, dconn2, struc_slice)

        if struc_bm.is_surface.all():
            cortex.append(struc_slice)
        else:
            sub_cortical.append(struc_slice)
        if 'LEFT' in short_name:
            left.append(struc_slice)
        if 'RIGHT' in short_name:
            right.append(struc_slice)

    for name in ('cortex', 'sub_cortical', 'left', 'right'):
        logger.info('Computing correlation relative to all %s brain structures', name)
        res[name] = correlate(dconn1, dconn2, locals()[name])

    return res


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    logger.info('starting %s', op.basename(__file__))
    dconn1 = cifti2.load(args.input)
    axes1 = dconn1.header.get_axis(dconn1)
    dconn2 = cifti2.load(args.reference)
    axes2 = dconn2.header.get_axis(dconn2)
    if not (isinstance(axes1[1], cifti2.BrainModelAxis) or isinstance(axes1[1], cifti2.ParcelsAxis)):
        raise ValueError("Columns should have greyordinates or parcels")
    if axes1[1] != axes2[1]:
        raise ValueError("Compared CIFTI files should have the same greyordinates/parcels along the columns")
    if args.as_dconn:
        if isinstance(axes1[0], cifti2.SeriesAxis):
            axes1 = (axes1[1], ) * 2
            dconn1 = FakeDConn(dconn1)
        elif isinstance(axes2[0], cifti2.SeriesAxis):
            axes2 = (axes2[1], ) * 2
            dconn2 = FakeDConn(dconn2)
    if args.split_dconn and not isinstance(axes1[0], cifti2.BrainModelAxis):
        raise ValueError("Rows should be greyordinates when using the option --split_dconn")
    if axes1[0] != axes2[0]:
        raise ValueError("Compared CIFTI files should have the same features along the rows")
    as_dict = run(
            dconn1=dconn1,
            dconn2=dconn2,
            split_greyordinates=axes1[0] if args.split_dconn else None,
    )
    scalar = cifti2.ScalarAxis(list(as_dict.keys()))
    arr = np.stack(list(as_dict.values()), -1)
    cifti2.Cifti2Image(arr.T, header=(scalar, axes1[1])).to_filename(args.output)
    logger.info('ending %s', op.basename(__file__))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument("-i", "--input", required=True,
                        help="input parcellated or dense CIFTI file")
    parser.add_argument("-r", "--reference", required=True,
                        help="reference dense connectome (should have the same greyordinates or parcels)")
    parser.add_argument("-o", "--output", required=True,
                        help="output dscalar or pscalar file with the correlations")
    parser.add_argument("--split_dconn", action='store_true',
                        help='If set splits the feature greyordinates into different groups (works for dconn or pdconn)')
    parser.add_argument("--as_dconn", action='store_true',
                        help="If set treats dense timeseries files and dense connectome files " +
                             "using the Pearson r correlation")



import argparse
import os
import shutil
from contextlib import contextmanager
from typing import Sequence

import dask.array
import nibabel as nib
import numpy as np
from fsl.utils.path import hasExt
from nibabel.cifti2 import Cifti2Header, Cifti2Image
from nibabel.cifti2.cifti2_axes import Axis, BrainModelAxis, ScalarAxis
from nibabel.gifti import GiftiImage
from numpy.lib.stride_tricks import as_strided

from mcot.surface import cortical_mesh, write_gifti

from . import cifti


class GreyOrdinates(object):
    """Represents data on voxels or vertices."""

    def __init__(self, data, brain_model_axis: BrainModelAxis, other_axes: Sequence[Axis]=None, parent_file=None):
        """Defines a new dataset in greyordinate space.

        :param data: (..., N) array for N greyordinates
        :param brain_model_axis: CIFTI axis describing the greyordinate space
        :param other_axes: sequence of CIFTI axes describing the other dimensions
        :param parent_file: file in which the dataset has been stored
        """
        self.data = data
        if data.shape[-1] != len(brain_model_axis):
            raise ValueError("Last axis of data does not match number of greyordinates")
        self.brain_model_axis = brain_model_axis
        if other_axes is not None:
            if len(other_axes) != self.data.ndim - 1:
                raise ValueError("Number of axis does not match dimensionality of the data")
            if tuple(len(ax) for ax in other_axes) != self.data.shape[:-1]:
                raise ValueError("Size of other axes does not match data size")
        self.other_axes = None if other_axes is None else tuple(other_axes)
        self.parent_file = parent_file

    def volume(self, ):
        """Get the volumetric data as a Nifti1Image."""
        if self.brain_model_axis.volume_mask.sum() == 0:
            raise ValueError(f"Can not create volume without voxels in {self}")
        data = np.full(self.brain_model_axis.volume_shape + self.data.shape[:-1], np.nan,
                       dtype=self.data.dtype)
        voxels = self.brain_model_axis.voxel[self.brain_model_axis.volume_mask]
        data[tuple(voxels.T)] = np.transpose(self.data, (-1, ) + tuple(range(self.data.ndim - 1)))[self.brain_model_axis.volume_mask]
        return nib.Nifti1Image(data, affine=self.brain_model_axis.affine)

    def surface(self, anatomy, fill=np.nan, partial=False):
        """Gets a specific surface.

        :param anatomy: BrainStructure or string like 'CortexLeft' or 'CortexRight'
        :param fill: which value to fill the array with if not all vertices are defined
        :param partial: only return the part of the surface defined in the greyordinate file (ignores `fill` if set)
        :return:
            - if not partial: (..., n_vertices) array
            - if partial: (N, ) int array with indices on the surface included in (..., N) array
        """
        if isinstance(anatomy, str):
            anatomy = cortical_mesh.BrainStructure.from_string(anatomy, issurface=True)
        if anatomy.cifti not in self.brain_model_axis.name:
            raise ValueError(f"No surface data for {anatomy.cifti} found")
        slc, bm = None, None
        arr = np.full(self.data.shape[:-1] + (self.brain_model_axis.nvertices[anatomy.cifti],), fill,
                      dtype=self.data.dtype)
        for name, slc_try, bm_try in self.brain_model_axis.iter_structures():
            if name == anatomy.cifti:
                if partial:
                    if bm is not None:
                        raise ValueError(f"Surface {anatomy} does not form a contiguous block")
                    slc, bm = slc_try, bm_try
                else:
                    arr[..., bm_try.vertex] = self.data[..., slc_try]
        if not partial:
            return arr
        else:
            return bm.vertex, self.data[..., slc]

    def to_hdf5(self, group, compression='gzip'):
        """Stores the image in the HDF5 group."""
        other_axes = (None, ) * (self.data.ndim - 1) if self.other_axes is None else self.other_axes
        cifti.to_hdf5(group, self.data, other_axes + (self.brain_model_axis, ), compression=compression)

    @classmethod
    def from_hdf5(cls, group):
        """Retrieves data from HDF5 group."""
        data, axes = cifti.from_hdf5(group)
        return cls(data, axes[-1], axes[:-1])

    @classmethod
    @contextmanager
    def empty_hdf5(cls, filename, axes, dtype=float):
        """Creates an empty greyordinate object based on the axes.

        Data will be stored on disk in an HDF5 file

        :param filename: filename to store the HDF5 file in or HDF5 group
        :param axes: cifti2 axes
        :param dtype: data type of array
        :return: Greyordinate object where the data is stored in the new HDF5 file
        """
        import h5py
        if isinstance(filename, str):
            group = h5py.File(filename, 'w')
            to_close = True
        else:
            group = filename
            to_close = False
        data = cifti.empty_hdf5(group, axes, dtype)
        try:
            yield cls(data, axes[-1], axes[:-1], parent_file=group)
        except:
            if to_close:
                os.remove(filename)
            raise
        if to_close:
            group.close()

    @classmethod
    @contextmanager
    def empty_zarr(cls, filename, axes, dtype=float):
        """Creates an empty greyordinate object based on the axes.

        Data will be stored on disk in an HDF5 file

        :param filename: filename to store the HDF5 file in or HDF5 group
        :param axes: cifti2 axes
        :param dtype: data type of array
        :return: Greyordinate object where the data is stored in the new HDF5 file
        """
        import zarr
        if isinstance(filename, str):
            group = zarr.group(filename, 'w')
        else:
            group = filename
        data = cifti.empty_zarr(group, axes, dtype)
        try:
            yield cls(data, axes[-1], axes[:-1], parent_file=group)
        except:
            to_delete = filename if group is not filename else group.path
            if os.path.isdir(to_delete):
                shutil.rmtree(to_delete)
            raise

    def to_cifti(self, other_axes=None):
        """Create a CIFTI image from the data.

        :param other_axes: defines the axes besides the greyordinate one
        :return: nibabel CIFTI image
        """
        if other_axes is None:
            other_axes = self.other_axes
            if other_axes is None:
                if self.data.ndim != 1:
                    raise ValueError("Can not store to CIFTI without defining what is stored along the other dimensions")
                other_axes = []
        if other_axes is not None:
            if len(other_axes) != self.data.ndim - 1:
                raise ValueError("Number of axis does not match dimensionality of the data")
            if tuple(len(ax) for ax in other_axes) != self.data.shape[:-1]:
                raise ValueError("Size of other axes does not match data size")

        data = self.data
        if data.ndim == 1:
            data = data[None, :]
            other_axes = [ScalarAxis(['default'])]

        return Cifti2Image(
                data,
                header=Cifti2Header.from_axes(list(other_axes) + [self.brain_model_axis])
        )

    @classmethod
    def from_cifti(cls, filename, writable=False):
        """Creates new greyordinate object from dense CIFTI file.

        :param filename: CIFTI filename
        :param writable: if True, opens data array in writable mode
        """
        if isinstance(filename, str):
            img: Cifti2Image = nib.load(filename)
        else:
            img = filename
        if writable:
            data = np.memmap(filename, img.dataobj.dtype, mode='r+',
                             offset=img.dataobj.offset, shape=img.shape, order='F')
        else:
            data = np.asanyarray(img.dataobj)
        axes = [img.header.get_axis(idx) for idx in range(data.ndim)]
        if not isinstance(axes[-1], BrainModelAxis):
            raise ValueError("Last axis of dense CIFTI file should be a BrainModelAxis")
        return GreyOrdinates(data, axes[-1], axes[:-1])

    @classmethod
    @contextmanager
    def empty_cifti(cls, filename, axes, dtype=float):
        """Creates an empty greyordinate object based on the axes.

        Data will be stored on disk in CIFTI format

        :param filename: filename to store the CIFTI file in
        :param axes: cifti2 axes
        :param dtype: data type of array
        :return: Greyordinate object where the data is stored in the new HDF5 file
        """
        hdr = cifti.cifti2.Cifti2Header.from_axes(axes)
        data = np.zeros(1, dtype=dtype)
        shape = tuple(len(ax) for ax in axes)
        data_shaped = as_strided(data, shape=shape, strides=tuple(0 for _ in axes),
                                 writeable=False)
        cifti.cifti2.Cifti2Image(data_shaped, header=hdr).to_filename(filename)
        go = cls.from_cifti(filename, writable=True)
        try:
            yield go
        except:
            os.remove(filename)
            raise
        go.data.flush()
        del go

    @classmethod
    def from_gifti(cls, filename, mask_values=(0, np.nan)):
        """Creates a new greyordinate object from a GIFTI file.

        :param filename: GIFTI filename
        :param mask_values: values to mask out
        :return: greyordinate object representing the unmasked vertices
        """
        if isinstance(filename, str):
            img = nib.load(filename)
        else:
            img = filename
        datasets = [darr.data for darr in img.darrays]
        if len(datasets) == 1:
            data = datasets[0]
        else:
            data = np.concatenate(
                    [np.atleast_2d(d) for d in datasets], axis=0
            )
        mask = np.ones(data.shape, dtype='bool')
        for value in mask_values:
            if value is np.nan:
                mask &= ~np.isnan(data)
            else:
                mask &= ~(data == value)
        while mask.ndim > 1:
            mask = mask.any(0)

        anatomy = cortical_mesh.get_brain_structure(img)

        bm_axes = BrainModelAxis.from_mask(mask, name=anatomy.cifti)
        return GreyOrdinates(data[..., mask], bm_axes)

    @classmethod
    def from_nifti(cls, filename, mask_values=(np.nan, 0)):
        """Creates a new greyordinate object from a NIFTI file.

        :param filename: NIFTI filename
        :param mask_values: which values to mask out
        :return: greyordinate object representing the unmasked voxels
        """
        if isinstance(filename, str):
            img = nib.load(filename)
        else:
            img = filename
        data = img.get_fdata()

        mask = np.ones(data.shape, dtype='bool')
        for value in mask_values:
            if value is np.nan:
                mask &= ~np.isnan(data)
            else:
                mask &= ~(data == value)
        while mask.ndim > 3:
            mask = mask.any(-1)

        inverted_data = np.transpose(data[mask], tuple(range(1, data.ndim - 2)) + (0, ))
        bm_axes = BrainModelAxis.from_mask(mask, affine=img.affine)
        return GreyOrdinates(inverted_data, bm_axes)

    def __add__(self, other):
        """Adds the overlapping part of the arrays.

        Only voxels/vertices in common between the greyordinate spaces
        are added
        """
        if not isinstance(other, GreyOrdinates):
            return NotImplemented

        if self.other_axes is not None:
            if other.other_axes is not None:
                if self.other_axes != other.other_axes:
                    raise ValueError("can not concatenate greyordinates when non-brain-model axes do not match")
            other_axes = self.other_axes
        else:
            other_axes = other.other_axes

        new_bm, slices = cifti.combine([self.brain_model_axis, other.brain_model_axis])

        return GreyOrdinates(
                data=self.data[..., slices[0]] + other.data[..., slices[1]],
                brain_model_axis=new_bm,
                other_axes=other_axes,
        )

    @classmethod
    @contextmanager
    def empty(cls, filename, axes, dtype=float):
        """Creates an empty file to store the greyordinates with the type
        determined by the extension:

        - .nii: CIFTI file
        - .h5/hdf5/he2/he5: HDF5 file representing CIFTI data
        - .zarr: zarr file representing CIFTI data

        :param filename: target filename
        :param axes: cifti2 axes
        :param dtype: data type of array
        :return: Greyordinate object where CIFTI data can be stored
        """
        formats = {
            'zarr': ('.zarr',),
            'hdf5': ('.hdf5', '.h5', '.he2', '.he5'),
            'cifti': ('.nii',),
        }
        for format, exts in formats.items():
            if hasExt(filename, exts):
                break
        else:
            raise ValueError(f"Extension of {filename} not recognised as CIFTI, HDF5 or zarr file")
        with getattr(cls, f'empty_{format}')(filename, axes, dtype) as go:
            yield go

    @classmethod
    def from_filename(cls, filename, mask_values=(0, np.nan), writable=False):
        """Reads greyordinate data from the given file.

        File can be:

        - NIFTI mask
        - GIFTI mask
        - CIFTI file
        - HDF5 file representing CIFTI data
        - zarr file representing CIFTI data

        :param filename: input filename
        :param mask_values: which values are outside of the mask for NIFTI or GIFTI input
        :param writable: allow to write to disk
        :return: greyordinates object
        """
        try:
            import h5py
        except ImportError:
            h5py = False
        try:
            import zarr
        except ImportError:
            zarr = False
        if isinstance(filename, str):
            try:
                if not h5py:
                    raise OSError()
                img = h5py.File(filename, 'r+' if writable else 'r')
            except OSError:
                try:
                    if not zarr:
                        raise ValueError()
                    img = zarr.open(filename, mode='r+' if writable else 'r')
                except ValueError:
                    img = nib.load(filename)
        else:
            img = filename

        if h5py and isinstance(img, h5py.Group):
            return cls.from_hdf5(img)
        if zarr and isinstance(img, zarr.Group):
            return cls.from_hdf5(img)
        if isinstance(img, nib.Nifti1Image):
            if writable:
                raise ValueError("Can not open NIFTI file in writable mode")
            return cls.from_nifti(filename, mask_values)
        if isinstance(img, Cifti2Image):
            return cls.from_cifti(filename, writable=writable)
        if isinstance(img, GiftiImage):
            if writable:
                raise ValueError("Can not open GIFTI file in writable mode")
            return cls.from_gifti(filename, mask_values)
        raise ValueError(f"I do not know how to convert {type(img)} into greyordinates (from {filename})")

    def to_filename(self, filename):
        """Stores the greyordinate data to the given filename.

        Type of storage is determined by the extension of the filename:

        - .dscalar/dconn/dlabel.nii: CIFTI file
        - .h5/hdf5/he2/he5: HDF5 file representing CIFTI data
        - .zarr: zarr file representing CIFTI data
        - .gii: GIFTI file (only stores surface data;
            raises error if more that one surface is represented in the greyordinates)
        - .nii: NIFTI file (only stores the volumetric data)

        :param filename: target filename
        """
        if hasExt(filename, ('.dscalar.nii', '.dconn.nii', '.dlabel.nii')):
            self.to_cifti().to_filename(filename)
        elif hasExt(filename, ('.h5', '.hdf5', '.he2', 'he5')):
            import h5py
            with h5py.File(filename, 'w') as f:
                self.to_hdf5(f)
        elif hasExt(filename, ('.zarr', )):
            import zarr
            f = zarr.group(filename)
            self.to_hdf5(f)
        elif hasExt(filename, ('.gii', )):
            surfaces = np.unique(self.brain_model_axis.name[self.brain_model_axis.surface_mask])
            if len(surfaces) > 1:
                raise ValueError(f"Can not write to GIFTI file as more than one surface has been defined: {surfaces}")
            if len(surfaces) == 0:
                raise ValueError("Can not write to GIFTI file as no surface has been provided")
            write_gifti(filename, [self.surface(surfaces[0])], surfaces[0])
        elif hasExt(filename, ('.nii.gz', '.nii')):
            self.volume().to_filename(filename)
        else:
            raise IOError(f"Extension of {filename} not recognized for NIFTI, GIFTI, or CIFTI file")

    def as_dask(self, chunks='auto', name='greyordinates'):
        """Returns the greyordinates as a dask array.

        :param chunks: shape of the chunks (defaults to chunks of the dataset)
        :param name: name of the dask array
        :return: dask array
        """
        if chunks == 'auto':
            chunks = getattr(self.data, 'chunks', 'auto')
            if chunks != 'auto':
                size_chunks = np.prod(chunks)
                if size_chunks < 2e7:
                    mult_factor = int((2e7 / size_chunks) ** (1. / self.data.ndim))
                    chunks = tuple(c * mult_factor for c in chunks)
        return dask.array.from_array(self.data, chunks, name)

    def transpose(self, ):
        """Transposes a dense connectome."""
        if self.data.ndim != 2:
            raise ValueError("Can only transpose 2D datasets")
        return GreyOrdinates(np.transpose(self.data), brain_model_axis=self.other_axes[0],
                             other_axes=(self.brain_model_axis, ), parent_file=self.parent_file)


def stack(greyordinates, axis=0):
    """Stacks a sequene of greyordinates along the given axis.

    Resulting GreyOrdinate will only contain voxels/vertices in all GreyOrdinate arrays

    :param greyordinates: individual greyordinates to be merged
    :return: merged greyordinate object
    """
    new_bm, slices = cifti.combine([go.brain_model_axis for go in greyordinates])
    new_arr = np.stack([go.data[..., slc] for go, slc in zip(greyordinates, slices)], axis=axis)
    ref_axes = set([go.other_axes for go in greyordinates if go.other_axes is not None])
    if len(ref_axes) == 0:
        other_axes = None
    elif len(ref_axes) == 1:
        other_axes = list(ref_axes[0])
        other_axes.insert(axis, ScalarAxis([f'stacked_{idx + 1}' for idx in range(len(greyordinates))]))
    else:
        raise ValueError("Failed to merge greyordinates as their other axes did not match")
    return GreyOrdinates(new_arr, new_bm, other_axes)


def concatenate(greyordinates, axis=0):
    """Stacks a sequene of greyordinates along the given axis.

    Resulting GreyOrdinate will only contain voxels/vertices in all GreyOrdinate arrays

    :param greyordinates: individual greyordinates to be merged
    :return: merged greyordinate object
    """
    if len(greyordinates) == 1:
        return greyordinates[0]
    ref_axes = [go.other_axes for go in greyordinates if go.other_axes is not None]
    if len(ref_axes) == 0:
        other_axes = None
    elif any(ra != ref_axes[0] for ra in ref_axes[1:]):
        raise ValueError("Failed to merge greyordinates as their other axes did not match")
    else:
        other_axes = ref_axes[0]

    if axis == -1 or axis == greyordinates[0].ndim - 1:
        new_bm = greyordinates[0].brain_model_axis
        for go in greyordinates[1:]:
            new_bm = new_bm + go.brain_model_axis
        new_arr = np.concatenate([go.data for go in greyordinates], axis=axis)
        return GreyOrdinates(new_arr, new_bm, other_axes)
    else:
        new_bm, slices = cifti.combine([go.brain_model_axis for go in greyordinates])
        new_arr = np.concatenate([go.data[..., slc] for go, slc in zip(greyordinates, slices)], axis=axis)
        return GreyOrdinates(new_arr, new_bm, other_axes)


def parse_greyordinate(filename):
    """Parses a set of filenames as a single greyordinate object.

    :param filename: '@'-symbol separated files (NIFTI, GIFTI, and/or CIFTI)
    :return: single Greyordinate object representing the full dataset
    """
    try:
        parts = [GreyOrdinates.from_filename(fn) for fn in filename.split('@')]
    except (ValueError, IOError) as e:
        raise argparse.ArgumentTypeError(*e.args)

    if len(parts) == 0:
        raise argparse.ArgumentParser("Can not parse empty string")

    return concatenate(parts, axis=-1)



import argparse
import os.path as op

import nibabel as nib
import numpy as np
from nibabel import cifti2, gifti
from nibabel.filebasedimages import ImageFileError

from mcot.surface._write_gifti import write_gifti

from ..surface.cortical_mesh import (BrainStructure, CorticalMesh,
                                     get_brain_structure)


def _nifti2cifti(img):
    arr = img.get_data()
    mask = arr != 0
    while mask.ndim > 3:
        mask = mask.any(-1)
    arr = arr[mask]
    bm = cifti2.BrainModelAxis.from_mask(mask, affine=img.affine)
    assert len(bm) == arr.shape[0]
    axes = tuple(cifti2.SeriesAxis(0, 1, sz) for sz in arr.shape[1:]) + (bm, )
    return arr.T, axes


def _gifti2cifti(left, right, table=None):
    if left is not None and left.size != 0:
        mask_left = np.isfinite(left) & (left != 0)
        while mask_left.ndim > 1:
            mask_left = mask_left.any(-1)
        bm_left = cifti2.BrainModelAxis.from_mask(mask_left, 'CortexLeft')

    if right.size != 0:
        mask_right = np.isfinite(right) & (right != 0)
        while mask_right.ndim > 1:
            mask_right = mask_right.any(-1)
        bm_right = cifti2.BrainModelAxis.from_mask(mask_right, 'CortexRight')

        if left.size == 0:
            bm = bm_right
            farr = right[mask_right]
        else:
            bm = bm_left + bm_right
            farr = np.concatenate((left[mask_left], right[mask_right]), 0)
    elif left.size != 0:
        bm = bm_left
        farr = left[mask_left]
    else:
        raise ValueError("Neither left nor right hemisphere provided")

    farr = np.squeeze(farr)
    if farr.ndim == 1:
        farr = farr[:, None]

    new_axes = tuple(range(1, farr.ndim)) + (0, )
    if table is None:
        axes = tuple(cifti2.ScalarAxis(['???'] * sz) for sz in farr.shape[1:]) + (bm, )
        return np.transpose(farr, new_axes), axes
    lab = cifti2.LabelAxis(['???'] * farr.shape[1], [table] * farr.shape[1])
    return np.transpose(farr, new_axes), (lab, bm)


def _cifti2nifti(arr, axes):
    bm = axes[-1]
    if not isinstance(bm, cifti2.BrainModelAxis):
        raise ValueError(f"CIFTI file should be dense to extract volumetric array")
    full_vol = np.zeros(bm.volume_shape + arr.shape[:-1])
    full_vol[tuple(bm.voxel[bm.volume_mask].T)] = arr[..., bm.volume_mask].T
    return nib.Nifti1Image(full_vol, bm.affine)


def _cifti2gifti(arr, axes, get_label=False):
    res = [np.array(()), np.array(())]
    bm = axes[-1]
    if not isinstance(bm, cifti2.BrainModelAxis):
        raise argparse.ArgumentTypeError(f"CIFTI file should be dense to extract surface array")
    for name, slc, sub_bm in bm.iter_structures():
        anatomy = BrainStructure.from_string(name)
        if anatomy == 'CortexLeft' or anatomy == 'CortexRight':
            full_arr = np.zeros((sub_bm.nvertices[name], ) + arr.shape[:-1])
            full_arr[()] = np.nan
            full_arr[sub_bm.vertex] = arr[..., slc].T
            res[anatomy == 'CortexRight'] = full_arr
    if not get_label:
        return tuple(res)
    label = axes[0]
    if not isinstance(label, cifti2.LabelAxis):
        raise ValueError("Input axis must be Label to extract parcellation label table")
    return tuple(res), label.label[0]


def greyordinate_in(value):
    """Input greyordinate image.

    3 types of input can be expected:
    - CIFTI file
    - GIFTI file
    - NIFTI file

    In the GIFTI and NIFTI files only non-zero values are kept

    :param value: input NIFTI, GIFTI, or CIFTI file
    :return: array, greyordinate axes
    """
    if '@' in value:
        arr = []
        for fn in value.split('@'):
            part_arr, axes = greyordinate_in(fn)
            arr.append(part_arr)
            if len(arr) == 1:
                ref_axes = axes[:-1]
                bm = axes[-1]
            else:
                assert axes[:-1] == ref_axes
                bm = bm + axes[-1]
        return np.concatenate(arr, -1), ref_axes + (bm, )

    if not op.isfile(value):
        raise argparse.ArgumentTypeError(f"Input greyordinate file {value} not found on disk")
    img = nib.load(value)
    try:
        img = nib.Cifti2Image.from_filename(value)
    except ImageFileError:
        pass
    if isinstance(img, nib.Cifti2Image):
        return np.asarray(img.dataobj), [img.header.get_axis(idx) for idx in range(img.ndim)]
    elif isinstance(img, gifti.GiftiImage):
        arr = np.squeeze(np.stack([darr.data for darr in img.darrays], 0))
        if img.labeltable is not None and len(img.labeltable.labels) > 1:
            table = {label.key: (label.label, label.rgba) for label in img.labeltable.labels}
        else:
            table = None
        if get_brain_structure(img).hemisphere == 'left':
            return _gifti2cifti(arr, np.array(()), table)
        else:
            return _gifti2cifti(np.array(()), arr, table)
    else:
        return _nifti2cifti(img)


def surface_arr_in(value):
    """Reads in arrays across the vertex.

    For the CIFTI files the output arrays are padded with zeros

    :param value: input GIFTI or CIFTI file
    :return: tuple with vertex values in left and right cortex
    """
    if not op.isfile(value):
        raise argparse.ArgumentTypeError(f"Input surface array file {value} not found on disk")
    img = nib.load(value)
    try:
        img = nib.Cifti2Image.from_filename(value)
    except ImageFileError:
        pass
    if isinstance(img, nib.Cifti2Image):
        arr, axes = np.asarray(img.dataobj), [img.header.get_axis(idx) for idx in range(img.ndim)]
        res = _cifti2gifti(arr, axes)
        if res[0].size == 0 and res[1].size == 0:
            raise argparse.ArgumentTypeError(f"CIFTI file {value} does not contain cortical surface")
        return res
    elif isinstance(img, gifti.GiftiImage):
        res = [np.zeros(()), np.zeros(())]
        arr = np.squeeze(np.stack([darr.get_data() for darr in img.darrays], -1))
        res[get_brain_structure(img) == 'CortexRight'] = arr
        return tuple(res)
    else:
        raise ValueError(f"Surface arrays should be stored in GIFTI or CIFTI files, not {type(img)}")


def surface_label_in(value):
    """Reads in surface parcellation from a label file.

    :param value: input GIFTI or CIFTI filename
    :return: tuple with vertex value and table in left and right cortex
    """
    if not op.isfile(value):
        raise argparse.ArgumentTypeError(f"Input surface label file {value} not found on disk")
    img = nib.load(value)
    try:
        img = nib.Cifti2Image.from_filename(value)
    except ImageFileError:
        pass
    if isinstance(img, nib.Cifti2Image):
        arr, axes = np.asarray(img.dataobj), [img.header.get_axis(idx) for idx in range(img.ndim)]
        res, table = _cifti2gifti(arr, axes, get_label=True)
        if res[0].size == 0 and res[1].size == 0:
            raise argparse.ArgumentTypeError(f"CIFTI file {value} does not contain cortical surface")
        return (res[0], table), (res[1], table)
    elif isinstance(img, gifti.GiftiImage):
        res = [(np.zeros(()), None), (np.zeros(()), None)]
        arr = np.squeeze(np.stack([darr.get_data() for darr in img.darrays], -1))
        table = {label.key: (label.label, label.rgba) for label in img.labeltable.labels}
        res[get_brain_structure(img) == 'CortexRight'] = (arr, table)
        return tuple(res)
    else:
        raise ValueError(f"Surface arrays should be stored in GIFTI or CIFTI files, not {type(img)}")


def volume_in(value):
    """Reads in a volume from a NIFTI or CIFTI file.

    :param value: input filename (NIFTI or CIFTI)
    :return: nibabel NIFTI1Image (or other volumetric image)
    """
    if not op.isfile(value):
        raise argparse.ArgumentTypeError(f"Input volume file {value} not found on disk")
    img = nib.load(value)
    try:
        img = nib.Cifti2Image.from_filename(value)
    except ImageFileError:
        pass
    if isinstance(value, nib.Cifti2Image):
        arr, axes = np.asarray(img.dataobj), [img.get_axis(idx) for idx in range(img.ndim)]
        return _cifti2nifti(arr, axes)
    else:
        return img


def surface_in(value):
    """Reads one or two surfaces from GIFTI files.

    To read both surfaces seperate them with an @ sign

    :param value: input filename (or '@'-separated filenames)
    :return: tuple with left and right surface
    """
    if '@' in value:
        fn1, fn2 = value.split('@')
        res1 = surface_in(fn1)
        res2 = surface_in(fn2)
        if res1[0] is not None and res2[0] is not None:
            raise argparse.ArgumentTypeError(f"Both surface files in {value} provide a left hemisphere")
        if res1[1] is not None and res2[1] is not None:
            raise argparse.ArgumentTypeError(f"Both surface files in {value} provide a right hemisphere")
        if res1[0] is None:
            return res2[0], res1[1]
        else:
            return res1[0], res2[1]
    else:
        if not op.isfile(value):
            raise argparse.ArgumentTypeError(f"Input surface file {value} not found on disk")
        surface = CorticalMesh.read(value)
        if surface.anatomy.hemisphere == 'left':
            return surface, None
        elif surface.anatomy.hemisphere == 'right':
            return None, surface
        else:
            raise argparse.ArgumentTypeError(f"Unrecognized hemisphere {surface.anatomy.hemisphere} in {value}")


def output(path, format=None):
    """Creates function to write provided output to a path.

    :param path: output filename
    :param format: Format of the output file. Default is based on extension:

        - '.gii' -> GIFTI (use '@'-separator to store left and right hemisphere separately)
        - '.nii' -> CIFTI
        - '.nii.gz' -> NIFTI

    :return: function writing a volume, surface, or greyordinate array
    """
    if format is None:
        if path[-4:] == '.gii':
            format = 'GIFTI'
        elif path[-4:] == '.nii':
            format = 'CIFTI'
        elif path[-7:] == '.nii.gz':
            format = 'NIFTI'
        else:
            raise argparse.ArgumentTypeError(f"Extension of output filename {path} not recognized")
    if format not in ('GIFTI', 'CIFTI', 'NIFTI'):
        raise argparse.ArgumentTypeError(f"Extension format {format} not recognized")

    def writer(obj):
        """Writes given object to the file.

        Can be one of:
        - Nifti1Image (for NIFTI/CIFTI output)
        - array/axes tuple (for any output)
        - tuple with 2 surface arrays (for GIFTI/CIFTI output)

        :param obj: input object
        """
        if not isinstance(obj, tuple):
            # NIFTI Image
            if format == 'GIFTI':
                raise ValueError(f"Can not write volumetric image {obj} to GIFTI path {path}")
            elif format == 'NIFTI':
                obj.to_filename(path)
            else:
                arr, axes = _nifti2cifti(obj)
                cifti2.Cifti2Image(arr, header=axes).to_filename(path)
            return

        part1, part2 = obj
        if isinstance(part2, np.ndarray):
            # surface arrays
            if format == 'NIFTI':
                raise ValueError(f"Can not write surface array to NIFTI path {path}")
            elif format == 'GIFTI':
                if part1.size == 0 and part2.size == 0:
                    raise ValueError("Only empty surface arrays provided")
                if part1.size == 0 or part2.size == 0:
                    if '@' in path:
                        raise ValueError("Single surface array provided to 2 GIFTI files")
                    if part1.size == 0:
                        write_gifti(path, [part2], 'CortexRight')
                    else:
                        write_gifti(path, [part1], 'CortexLeft')
                else:
                    fn1, fn2 = path.split('@')
                    write_gifti(fn1, [part1], 'CortexLeft')
                    write_gifti(fn2, [part2], 'CortexRight')
            else:
                arr, axes = _gifti2cifti(part1, part2)
                cifti2.Cifti2Image(arr, header=axes).to_filename(path)
        elif isinstance(part1, tuple):
            # surface labels
            arr1, table1 = part1
            arr2, table2 = part2
            if format == 'NIFTI':
                raise ValueError(f"Can not write surface array to NIFTI path {path}")
            elif format == 'GIFTI':
                if arr1.size == 0 and arr2.size == 0:
                    raise ValueError("Only empty surface arrays provided")
                if arr1.size == 0 or arr2.size == 0:
                    if '@' in path:
                        raise ValueError("Single surface array provided to 2 GIFTI files")
                    if arr1.size == 0:
                        write_gifti(path, [arr2], 'CortexRight', color_map=table2)
                    else:
                        write_gifti(path, [arr1], 'CortexLeft', color_map=table1)
                else:
                    fn1, fn2 = path.split('@')
                    write_gifti(fn1, [arr1], 'CortexLeft', color_map=table1)
                    write_gifti(fn2, [arr2], 'CortexRight', color_map=table2)
            else:
                if table1 != table2:
                    raise ValueError("Label tables should be the same for both surfaces to " +
                                     "combine into single CIFTI file")
                arr, axes = _gifti2cifti(arr1, arr2, table=table1)
                cifti2.Cifti2Image(arr, header=axes).to_filename(path)
        else:
            if format == 'CIFTI':
                cifti2.Cifti2Image(part1, header=part2).to_filename(path)
            elif format == 'NIFTI':
                _cifti2nifti(part1, part2).to_filename(path)
            else:
                try:
                    (a1, a2), table = _cifti2gifti(part1, part2, get_label=True)
                    as_gifti = ((a1, table), (a2, table))
                except ValueError:
                    as_gifti = _cifti2gifti(part1, part2)
                writer(as_gifti)
    return writer



#!/usr/bin/env python
"""Combines two parcellations into a single parcellation by multiplying
them."""
import argparse
from copy import deepcopy

import numpy as np

from mcot.cifti import cmd
from mcot.cifti.cifti import combine


def run(arr1, label1, arr2, label2):
    """Combines the two parcellations into one by multiplying.

    :param arr1: int array with first parcellation
    :param label1: dict from index to (name, RGBA colour) for first parcellation
    :param arr2: int array with second parcellation
    :param label2: dict from index to (name, RGBA colour) for second parcellation
    :return: tuple with two elements:

        - int array with final parcellation
        - dict from index to (name, RBBA colour)
    """
    arr2_values = np.array(list(label2.keys()), dtype='i4')
    frange2 = max(arr2_values.max() - arr2_values.min(), 1)
    factor = 10 ** int(np.ceil(np.log10(frange2)))
    combined = arr1 * factor + arr2 - arr2_values.min()

    table = {}
    for index1, (name1, col1) in label1.items():
        for index2, (name2, col2) in label2.items():
            idx = index1 * factor + index2 - arr2_values.min()
            if idx in combined:
                table[idx] = (
                    f'{name1} & {name2}',
                    tuple((np.array(col1) + np.array(col2)) / 2)
                )
    return combined, table


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    arr1, axes1 = args.first_parcellation
    arr2, axes2 = args.second_parcellation

    bm, (idx1, idx2) = combine((axes1[-1], axes2[-1]))
    arr1 = arr1[..., idx1]
    arr2 = arr2[..., idx2]
    axes1 = axes1[:-1] + (bm, )
    axes2 = axes2[:-1] + (bm, )

    if arr1.shape != arr2.shape:
        raise ValueError("Input parcellations should have the same shape")

    labels = []
    res = []
    for a1, x1, a2, x2 in zip(arr1, axes1[0].label, arr2, axes2[0].label):
        combined_arr, combined_table = run(a1, x1, a2, x2)
        res.append(combined_arr)
        labels.append(combined_table)

    label_axis = deepcopy(axes1[0])
    label_axis.label = labels
    args.output((
            np.stack(res, 0),
            (label_axis, ) + axes1[1:]
    ))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    if parser is None:
        parser = __doc__
    if isinstance(parser, str):
        parser = argparse.ArgumentParser(parser)
    parser.add_argument("first_parcellation", type=cmd.greyordinate_in,
                        help='GIFTI or CIFTI parcellation file')
    parser.add_argument("second_parcellation", type=cmd.greyordinate_in,
                        help='GIFTI or CIFTI parcellation file')
    parser.add_argument("output", type=cmd.output,
                        help='output GIFTI or CIFTI parcellation file')



from typing import Sequence, Tuple, Union

import h5py
import numpy as np
import zarr
from fsl.utils import path
from nibabel import cifti2

Group = Union[h5py.Group, zarr.Group]


file_types = [
    (3001, ['.dconn.nii'], (cifti2.BrainModelAxis, cifti2.BrainModelAxis)),
    (3002, ['.dtseries.nii'], (cifti2.SeriesAxis, cifti2.BrainModelAxis)),
    (3003, ['.pconn.nii'], (cifti2.ParcelsAxis, cifti2.ParcelsAxis)),
    (3004, ['.ptseries.nii'], (cifti2.SeriesAxis, cifti2.ParcelsAxis)),
    (3006, ['.dscalar.nii', '.dfan.nii'], (cifti2.ScalarAxis, cifti2.BrainModelAxis)),
    (3007, ['.dlabel.nii'], (cifti2.LabelAxis, cifti2.BrainModelAxis)),
    (3008, ['.pscalar.nii'], (cifti2.ScalarAxis, cifti2.ParcelsAxis)),
    (3009, ['.pdconn.nii'], (cifti2.BrainModelAxis, cifti2.ParcelsAxis)),
    (3010, ['.dpconn.nii'], (cifti2.ParcelsAxis, cifti2.BrainModelAxis)),
    (3011, ['.pconnseries.nii'], (cifti2.ParcelsAxis, cifti2.ParcelsAxis, cifti2.SeriesAxis)),
    (3012, ['.pconnscalar.nii'], (cifti2.ParcelsAxis, cifti2.ParcelsAxis, cifti2.ScalarAxis)),
]


def guess_extension(axes: Tuple[cifti2.Axis]) -> Sequence[str]:
    """Guesses the extension based on the CIFTI axes.

    :param axes: CIFTI axes describing the rows/columns of a CIFTI file
    :return: tuple of possible file extensions
    """
    for _, extensions, axes_types in file_types:
        if len(axes_types) == len(axes) and all(isinstance(a, at) for a, at in zip(axes, axes_types)):
            return tuple(extensions)
    return ()


def write(filename: str, arr: np.ndarray, axes: Tuple[cifti2.Axis]):
    """Writes a CIFTI file guessing the extension of the filename.

    :param filename: full filename of basename
    :param arr: array to be stored
    :param axes: CIFTI axes describing the rows/columns of a CIFTI file
    """
    extensions = guess_extension(axes)
    if len(extensions) == 0:
        raise ValueError("No valid extensions found for axes of type {}".format(type(a) for a in axes))
    new_filename = path.addExt(filename, allowedExts=extensions, mustExist=False, defaultExt=extensions[0])
    cifti2.Cifti2Image(arr, axes).to_filename(new_filename)


def _greyordinate_index(brain_model: cifti2.BrainModelAxis):
    assert (brain_model.name[0] == brain_model.name).all()
    idx = np.zeros(brain_model.size, dtype='i4')
    if not brain_model.surface_mask.all():
        voxel = brain_model.voxel[~brain_model.surface_mask]
        idx[brain_model.volume_mask] = brain_model.volume_shape[0] * (
            brain_model.volume_shape[1] * voxel[:, 2] +
            voxel[:, 1]
        ) + voxel[:, 0]
    idx[brain_model.surface_mask] = -brain_model.vertex[brain_model.surface_mask]
    return idx


def _find_overlap(bm1: cifti2.BrainModelAxis, bm2: cifti2.BrainModelAxis):
    full_idx1 = []
    full_idx2 = []
    as_dict1 = {n: (i, bm) for n, i, bm in bm1.iter_structures()}
    for name, idx2, bm2_part in bm2.iter_structures():
        if name in as_dict1:
            idx1, bm1_part = as_dict1[name]
            _, sub_idx1, sub_idx2 = np.intersect1d(
                    _greyordinate_index(bm1_part),
                    _greyordinate_index(bm2_part),
                    return_indices=True
            )
            sorter = np.argsort(sub_idx1)
            full_idx1.append(idx1.start + sub_idx1[sorter])
            full_idx2.append(idx2.start + sub_idx2[sorter])
    full_idx1 = np.concatenate(full_idx1, 0)
    full_idx2 = np.concatenate(full_idx2, 0)
    assert (len(full_idx1) == 0 and len(full_idx2) == 0) or (bm1[full_idx1] == bm2[full_idx2])
    return full_idx1, full_idx2


def combine(brain_models: Sequence[cifti2.BrainModelAxis]):
    """Find the common space of multiple BrainModel axes.

    :param brain_models: sequence of brain model axes
    :return: tuple of common brain model and sequence of indices with the common space
    """
    common_bm = brain_models[0]
    for bm in brain_models[1:]:
        common_bm = common_bm[_find_overlap(common_bm, bm)[0]]
    return common_bm, [_find_overlap(bm, common_bm)[0] for bm in brain_models]


def axis_from_hdf5(group: Group):
    """Stores the information from an axis in HDF5 group."""
    name = group.attrs['name']
    if name == 'None':
        return None
    if name == 'Scalar':
        return cifti2.ScalarAxis(np.array(group['name']).astype('U'))
    if name == 'BrainModel':
        nvertices = {str(key): int(value) for key, value in
                     zip(np.array(group.attrs['nvertices_keys']).astype('U'),
                         group.attrs['nvertices_values'])}
        return cifti2.BrainModelAxis(
            np.array(group['name']).astype('U'), np.array(group['voxel']),
            np.array(group['vertex']), np.array(group['affine']) if 'affine' in group else None,
            tuple(int(sz) for sz in group['volume_shape']) if 'volume_shape' in group else None,
            nvertices)
    if name == 'Parcels':
        nvertices = {str(key): int(value) for key, value in
                     zip(np.array(group.attrs['nvertices_keys']).astype('U'),
                         group.attrs['nvertices_values'])}
        voxels = [np.array(group[f'voxels{idx}']) for idx in range(len(group['name']))]
        vertices = []
        for idx in range(len(group['name'])):
            res = {}
            for name in np.array(group.attrs['nvertices_keys']).astype('U'):
                fname = f'vertices{idx}_{name}'
                if fname in group:
                    res[name] = np.array(group[fname])
            vertices.append(res)
        return cifti2.ParcelsAxis(
            np.array(group['name']).astype('S'), voxels, vertices,
            np.array(group['affine']) if 'affine' in group else None,
            tuple(int(sz) for sz in group['volume_shape']) if 'volume_shape' in group else None,
            nvertices)
    if name == 'Series':
        return cifti2.SeriesAxis(
            group.attrs['start'], group.attrs['step'], group.attrs['size'], group.attrs['unit']
        )
    raise ValueError(f"Reading {name} from HDF5 is not currently supported")


def axis_to_hdf5(group: Group, axis: cifti2.Axis):
    """Stores the information from an axis in HDF5 group."""
    def convert(as_list):
        if isinstance(group, zarr.Group):
            return as_list
        return np.array(as_list).astype('S')

    if axis is None:
        group.attrs['name'] = 'None'
    elif isinstance(axis, cifti2.ScalarAxis):
        group.attrs['name'] = 'Scalar'
        group['name'] = axis.name.astype('S')
    elif isinstance(axis, cifti2.BrainModelAxis):
        group.attrs['name'] = 'BrainModel'
        group['name'] = axis.name.astype('S')
        group['voxel'] = axis.voxel
        group['vertex'] = axis.vertex
        if axis.affine is not None:
            group['affine'] = axis.affine
        if axis.volume_shape is not None:
            group['volume_shape'] = axis.volume_shape
        group.attrs['nvertices_keys'] = convert(list(axis.nvertices.keys()))
        group.attrs['nvertices_values'] = list(axis.nvertices.values())
    elif isinstance(axis, cifti2.ParcelsAxis):
        group.attrs['name'] = 'Parcels'
        for idx in range(len(axis)):
            for key, value in axis.vertices[idx].items():
                if key not in axis.nvertices:
                    raise KeyError(f"Defining vertices for undefined surface {key}")
                group[f'vertices{idx}_{key}'] = value
            group[f'voxels{idx}'] = axis.voxels[idx]
        group['affine'] = axis.affine
        group['name'] = axis.name.astype('S')
        group.attrs['nvertices_keys'] = convert(list(axis.nvertices.keys()))
        group.attrs['nvertices_values'] = list(axis.nvertices.values())
        group['volume_shape'] = axis.volume_shape
    elif isinstance(axis, cifti2.SeriesAxis):
        group.attrs['name'] = 'Series'
        group.attrs['start'] = axis.start
        group.attrs['size'] = axis.size
        group.attrs['step'] = axis.step
        group.attrs['unit'] = axis.unit
    else:
        raise ValueError(f"storing {axis.__class__.__name__} in HDF5 is not currently supported")


def from_hdf5(group: Group) -> Tuple["h5py.Dataset", Sequence[cifti2.Axis]]:
    """Reads a CIFTI array from the HDF5 format.

    :param group: HDF5 group the data was stored in
    :return: tuple with data array (still on disk) and sequence of axes
    """
    data = group['data']
    axes = [axis_from_hdf5(group[f'axis{idx}']) for idx in range(data.ndim)]
    return data, axes


def to_hdf5(group: Group, arr, axes: Sequence[cifti2.Axis], compression='gzip'):
    """Store the CIFTI array in an HDF5 format.

    :param group: HDF5 group to store the data in (can be top-level HDF5 file)
    :param arr: data array
    :param axes: sequence of axes (optionally None)
    :param compression: which compression to use on the main data array (None, 'gzip', or 'lzf')
    """
    assert len(axes) == arr.ndim
    assert all(ax is None or len(ax) == sz for sz, ax in zip(arr.shape, axes))
    group.create_dataset('data', data=arr, compression=compression)
    for idx, axis in enumerate(axes):
        axis_to_hdf5(group.create_group(f'axis{idx}'), axis)


def empty_hdf5(group: Group, axes: Sequence[cifti2.Axis], dtype=float, compression='gzip'):
    """Creates a new HDF5 group with an empty dataset.

    :param group: HDF5 group
    :param axes: sequence of axes (all have to be defined)
    :param dtype: data type
    :param compression: which compression to use on each chunk
    :return: new array to be filled
    """
    group.create_dataset('data', shape=(len(ax) for ax in axes), dtype=dtype, compression=compression)
    for idx, axis in enumerate(axes):
        axis_to_hdf5(group.create_group(f'axis{idx}'), axis)
    return group['data']


def empty_zarr(group: Group, axes: Sequence[cifti2.Axis], dtype=float, compressor='default'):
    """Creates a new zarr group with an empty dataset.

    :param group: zarr group
    :param axes: sequence of axes (all have to be defined)
    :param dtype: data type
    :param compressor: which compressor to use on each chunk
    :return: new array to be filled
    """
    group.create_dataset('data', shape=(len(ax) for ax in axes), dtype=dtype, compressor=compressor)
    for idx, axis in enumerate(axes):
        axis_to_hdf5(group.create_group(f'axis{idx}'), axis)
    return group['data']



#!/usr/bin/env python
"""Parcellates the surface based on clustering similar fingerprints."""
import numpy as np
from loguru import logger

from mcot.cifti.greyordinate import GreyOrdinates
from mcot.surface import CorticalMesh, write_gifti

from ..cifti import gradient
from ..parcel import cluster, random


def run(surface: CorticalMesh, features, n_iter=20, metric='spearman', method='spectral'):
    """Repeatedly identify the border between clusters in a spotlight fashion.

    :param surface: anatomical surface with N vertices
    :param features: (N, M) array of surface features
    :param n_iter: number of  iterations
    :param metric: which metric to use for the similarity between neighbouring vertices
    :param method: clustering method ('spectral', 'affinity', or 'DBSCAN')
    :return: (N, ) array with fraction of times a vertex was on the border between the clusters
    """
    sim_mat = gradient.distance(surface.graph_point_point(include_diagonal=False), features, metric=metric)
    on_border = np.zeros(surface.nvertices)
    for n_clusters in np.linspace(10, 100, n_iter):
        logger.info('Randomly parcellating cortex in %d clusters', n_clusters)
        parcellation = random.run(surface, int(n_clusters))
        logger.info('Running %s clustering on each cluster', method)
        for idx_cluster in range(int(n_clusters)):
            logger.debug('Processing cluster %d (out of %d)', idx_cluster, int(n_clusters))
            mask = parcellation == idx_cluster
            labels = cluster.cluster(sim_mat[mask, :][:, mask], method=method, n_clusters=2)
            graph = surface[mask].graph_point_point()
            smooth_labels = (graph.dot(labels) + labels) / (graph.dot(np.ones(mask.sum())) + 1)
            # edge detection only guaranteed to work properly for 2 clusters
            on_border[mask] += (smooth_labels % 1.) != 0.
    return on_border / n_iter


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    surf = CorticalMesh.read(args.surface)
    go = GreyOrdinates.from_filename(args.features)
    features = go.surface(surf.anatomy)
    mask = np.isfinite(features)
    res = run(
            surf[mask],
            features,
            n_iter=args.iter,
            metric=args.metric,
            method=args.cluster,
    )
    full_res = np.zeros(surf.nvertices)
    full_res[mask] = res
    write_gifti(args.output, [full_res], surf.anatomy)


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument("surface", help='.surf.gii file describing the surface (WM/GM boundary, mid, or pial)')
    parser.add_argument("features", help='GIFTI/CIFTI file with the features of interest')
    parser.add_argument("output", help='GIFTI file of the most consistent edge locations (same type as input)')
    parser.add_argument("--iter", type=int, default=20, help='number of searchlights for every vertex')
    parser.add_argument("-m", '--metric', choices=gradient.dist_options, default='euclidean',
                        help='how to compute the similarity with the neighbours (default: euclidean)')
    parser.add_argument("-c", '--cluster', choices=('spectral', 'affinity', 'dbscan'), default='spectral',
                        help='Clustering algorithm to use')



#!/usr/bin/env python
"""Smooths the values in the GIFTI or CIFTI file across the surface."""
import numpy as np
from loguru import logger
from nibabel import cifti2

from mcot.cifti import cmd
from mcot.surface.cortical_mesh import BrainStructure


def smooth(surface, array, width, vertices=None, axis=-1):
    """Smooths the array over the surface.

    :param surface: Mesh representing the cortical surface
    :param array: values on the vertices of the array
    :param width: width of the smoothing kernel (in mm)
    :param vertices: indices indicating which vertices have values (default: all vertices)
    :return:
    """
    if width <= 0:
        return array
    if axis < 0:
        axis = axis + array.ndim
    if vertices is not None:
        surface = surface[vertices]
    neighbour_distance = surface.graph_point_point('distance').tocoo()

    min_dist = np.median(neighbour_distance.tocoo().data) / 6
    nsteps = int(round(width ** 2 / min_dist ** 2))
    weight = neighbour_distance.copy()
    if nsteps <= 1:
        nsteps = 1
    start = width ** 2 / nsteps
    weight.data = np.append(
            np.exp(-neighbour_distance.data ** 2 / (2 * start)),
            np.ones(surface.nvertices)
    )
    weight.row = np.append(
            weight.row,
            np.arange(surface.nvertices)
    )
    weight.col = np.append(
            weight.col,
            np.arange(surface.nvertices)
    )
    weight = weight.tocsr()

    trans_axes = [axis] + list(range(axis)) + list(range(axis + 1, array.ndim))
    trans_array = np.transpose(array, trans_axes)
    summed = np.array(weight.sum(0)).flatten()[:, None]
    trans_reshaped = trans_array.reshape((trans_array.shape[0], -1))
    for _ in range(nsteps):
        trans_reshaped = weight.dot(trans_reshaped) / summed
    new_axes = list(range(1, array.ndim))
    new_axes.insert(axis, 0)
    transposed = np.transpose(trans_reshaped.reshape(trans_array.shape), new_axes)
    assert transposed.shape == array.shape
    return transposed


def smooth_cifti(arr, axes, surfaces, width, overwrite=False):
    """Smooths a CIFTI array across provided surfaces.

    :param arr: input array (overwritten if `overwrite` is True)
    :param axes: CIFTI axes for the array
    :param surfaces: tuple of left & right surface (None if not smoothing on that surface)
    :param width: width of the smoothing kernel (in mm)
    :param overwrite: if True overwrites the input array (saves memory)
    :return: smoothed array
    """
    if overwrite:
        res = arr
    else:
        res = arr.copy()

    smoothed = False
    for dim, bm in enumerate(axes):
        if isinstance(bm, cifti2.BrainModelAxis):
            selector = [slice(None)] * arr.ndim
            for name, slc, bm_part in bm.iter_structures():
                name = BrainStructure.from_string(name)
                selector[dim] = slc
                if name == 'CortexLeft' and surfaces[0] is not None:
                    logger.info(f'Smoothing left hemisphere across dimension {dim}')
                    smoothed = True
                    res[tuple(selector)] = smooth(surfaces[0], arr[tuple(selector)], width,
                                                  bm_part.vertex, axis=dim)
                elif name == 'CortexRight' and surfaces[1] is not None:
                    logger.info(f'Smoothing right hemisphere across dimension {dim}')
                    smoothed = True
                    res[tuple(selector)] = smooth(surfaces[1], arr[tuple(selector)], width,
                                                  bm_part.vertex, axis=dim)
    if not smoothed:
        raise ValueError("Provided surfaces did not match the hemispheres in the input file; no smoothing performed")
    return res


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    arr, axes = args.input
    res = smooth_cifti(arr, axes, args.surface, args.width, overwrite=True)
    args.output((res, axes))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument('input', type=cmd.greyordinate_in,
                        help='CIFTI or GIFTI file with the surface array (in a GIFTI file zeros are masked out)')
    parser.add_argument('surface', type=cmd.surface_in,
                        help=".surf.gii file with the surfaces (both can be provided if separated by '@'")
    parser.add_argument('width', type=float, help='smoothing kernel width in mm')
    parser.add_argument('output', type=cmd.output,
                        help='CIFTI or GIFTI file with the smoothed output (ignoring zeroes)')



#!/usr/bin/env python
"""Computes the feature similarity with the neighbouring vertices.

All relevant distance metric from scipy are included (https://docs.scipy.org/doc/scipy/reference/spatial.distance.html)
as well as histogram intersection (http://blog.datadive.net/histogram-intersection-for-change-detection/).
"""
from concurrent import futures

import numpy as np
from nibabel import cifti2
from scipy import sparse, spatial

from mcot.cifti.greyordinate import GreyOrdinates
from mcot.surface import CorticalMesh

scipy_options = (
    'braycurtis', 'canberra', 'cityblock', 'correlation',
    'cosine', 'euclidean', 'hamming', 'jaccard', 'jensenshannon',
    'minkowski', 'sqeuclidean',
)

dist_options = ('histogram_intersection', ) + scipy_options


def _histogram_intersection(arr1, arr2):
    normed1 = np.array(arr1, dtype='float')
    normed1 /= normed1.sum()
    normed2 = np.array(arr2, dtype='float')
    normed2 /= normed2.sum()
    return np.minimum(normed1, normed2).sum()


def distance(graph, features, metric='euclidean'):
    """Computes the distance between neighbouring feature sets.

    :param graph: (N, N) sparse matrix indicating which features are neighbouring
    :param features: (N, M) for M features defined at N locations
    :param metric: which distance operator to use (one of `dist_options`)
    :return: (N, N) sparse matrix with the distance between each feature pair (same structure as `graph`)
    """
    executor = futures.ThreadPoolExecutor(10)
    metric = metric.lower()
    if metric not in dist_options:
        raise ValueError(f"{metric} is not a valid distance metric, should be one of: {', '.join(dist_options)}")
    if metric == 'histogram_intersection':
        dist_func = _histogram_intersection
    else:
        dist_func = getattr(spatial.distance, metric)
    coo = graph.tocoo()

    data = np.zeros(coo.data.size)

    def to_run(idx):
        data[idx] = dist_func(features[coo.row[idx], :], features[coo.col[idx], :])
    list(executor.map(to_run, range(coo.data.size), chunksize=100))
    return sparse.coo_matrix((data, (coo.row, coo.col)), shape=graph.shape)


def reduce(distance, method='median'):
    """Summarises a NxN sparse matrix of distances as a (N, ) array of
    distances.

    :param distance: (N, N) sparse matrix of distances between neighbouring vertices
    :return: (N, ) array with the summarised distance
    """
    if method not in ('min', 'max', 'median', 'mean'):
        raise ValueError("Reduction method {method} has not been recognised")
    as_mat = distance.tocsr()
    res = np.zeros(distance.shape[0])
    for idx, i1, i2 in zip(range(distance.shape[0]), as_mat.indptr[:-1], as_mat.indptr[1:]):
        if i1 == i2:
            res[idx] = np.nan
        else:
            res[idx] = getattr(np, method)(as_mat.data[i1:i2])
    return res


def run(surface: CorticalMesh, arr, distance_metric='euclidean', reduction='median'):
    """Computes the reduced distance between neighbouring vertices.

    :param surface: surface array
    :param arr: (N, M) array for M features on N vertices
    :param distance_metric: distance metric to use
    :param reduction: reduction method to use to average over all connections with a vertex
    :return: (N, ) array with the summarised distance
    """
    if arr.shape[0] != surface.nvertices:
        raise ValueError(f"Expected ({surface.nvertices}, N) array, got {arr.shape} array")
    graph = surface.graph_point_point(dtype=float, include_diagonal=False)
    dist = distance(graph, arr, metric=distance_metric)
    return reduce(dist, method=reduction)


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    go = GreyOrdinates.from_filename(args.features)
    if args.transpose:
        go = go.transpose()
    surface = CorticalMesh.read(args.surface)
    surface_mask, features = go.surface(surface.anatomy, partial=True)
    if args.exclude_same_region:
        features = features[go.other_axes[0].name != surface.anatomy.cifti]
    res = run(surface[surface_mask], features.T, args.distance, args.reduction)
    new_bm = cifti2.BrainModelAxis.from_surface(surface_mask, nvertex=surface.nvertices, name=surface.anatomy.cifti)
    GreyOrdinates(res, new_bm).to_filename(args.output)


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument('features', help='GIFTI or CIFTI file with the features to select on')
    parser.add_argument('surface', help='surface file')
    parser.add_argument('output', help='GIFTI or CIFTI output file')
    parser.add_argument('-d', '--distance', choices=dist_options, default='euclidean',
                        help='distance metric to use (default: euclidean)')
    parser.add_argument('-r', '--reduction', choices=('min', 'max', 'median', 'mean'), default='median',
                        help='reduction method to summarise all the edges of one vertex (default: median)')
    parser.add_argument('-T', '--transpose', action='store_true',
                        help='Transpose the features array before extracting features (only makes sense for dconn)')
    parser.add_argument('--exclude-same-region', action='store_true',
                        help="in a dconn file do not include connections to the same cortex as features")



#!/usr/bin/env python
"""Discretizes a continuous variable."""
import colorcet as cc
import numpy as np
from loguru import logger

from mcot.cifti import cmd
from mcot.cifti.cifti import combine


def run_array(arr, nbins, bins=None, weight=None, include_zeros=False):
    """Returns a discretised version of the input array.

    :param arr: nibabel input image
    :param nbins: number of bins to extract
    :param bins: one of the following

        - None: use weight to set the bins
        - 'number': each parcel will have the same number of elements
        - 'regular': split the range from min to max in the input into equal bins
        - 1D array: explicit boundaries

    :param weight: selects the bins so each parcel has the same sum in this image (only used if bins is None)
    :param include_zeros: if True include zeros in the analysis
    :return: tuple with:

        - array with the parcels (zero where the original array was zero)
        - (nbins + 1, ...) array with the applied boundaries
    """
    if weight is not None and arr.shape[:weight.ndim] != weight.shape:
        raise ValueError("Shape of weight image does not match input image")

    res = np.zeros(arr.shape, dtype='i4')

    all_bins = np.zeros((nbins + 1, ) + (() if weight is None else arr.shape[weight.ndim:]))

    for idx in np.ndindex(*(() if weight is None else arr.shape[weight.ndim:])):
        sub_arr = arr[(Ellipsis, ) + idx]
        mask = slice(None) if include_zeros else sub_arr != 0

        if weight is not None:
            if (np.array(idx) == 0).all():
                logger.info('Using weight file to set bins')
            idx_sorted = np.argsort(sub_arr[mask])
            values = np.append(0, np.cumsum(weight[mask][idx_sorted]))
            edges = np.floor(np.interp(
                    np.linspace(0, values[-1], nbins + 1)[1:-1],
                    values,
                    np.arange(values.size),
            )).astype('int')
            use_bins_mid = sub_arr[mask][idx_sorted][edges]
            use_bins = np.append(-np.inf, np.append(use_bins_mid, np.inf))
        elif bins == 'regular':
            if (np.array(idx) == 0).all():
                logger.info('Using regularly spaced bins')
            use_bins = np.linspace(sub_arr.min(), sub_arr.max(), nbins + 1)
        elif bins == 'number':
            if (np.array(idx) == 0).all():
                logger.info('Setting bins to have identical number of elements in parcels')
            use_bins = np.sort(sub_arr[mask])[np.around(np.linspace(0, sub_arr[mask].size - 1, nbins + 1)).astype('i4')]
        else:
            use_bins = np.array(bins)
        assert use_bins.size == nbins + 1

        logger.debug(f'Bins for {idx}: {use_bins}')

        all_bins[(Ellipsis, ) + idx] = use_bins

        use_bins[-1] += 1e-8
        res[(Ellipsis, ) + idx][mask] = np.digitize(sub_arr[mask], use_bins)
    return res, all_bins


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    arr, axes = args.input
    bins, weight = None, None
    if args.equal_weight:
        weight, axes_weight = args.equal_weight
        bm, (idx_arr, idx_weight) = combine([axes[-1], axes_weight[-1]])
        axes = axes[:-1] + (bm, )
        arr = arr[..., idx_arr]
        weight = weight[..., idx_weight]
    elif args.equal_number:
        bins = 'number'
    elif args.equal_bin:
        bins = 'regular'
    elif args.set_bin:
        if len(args.set_bin) == args.nbins - 1:
            args.set_bin = np.append(-np.inf, np.append(args.set_bin, np.inf))
        assert len(args.set_bin) == args.nbins + 1
        bins = args.set_bin
    else:
        raise ValueError("No binning method selected")

    res, used_bins = run_array(
        arr, args.nbins, bins=bins, weight=weight, include_zeros=args.include_zeros
    )
    labels = [{int(idx): (f'{start:.2f} to {end:.2f}', c) for idx, start, end, c in zip(
            range(1, 100), used_bins[:-1], used_bins[1:], cc.glasbey)}]
    new_axes = (axes[0].to_label(labels), ) + axes[1:]
    args.output((res, new_axes))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument('input', type=cmd.greyordinate_in,
                        help='input NIFTI/GIFTI/CIFTI file')
    parser.add_argument('output', type=cmd.output,
                        help='output NIFTI/GIFTI/CIFTI files')
    parser.add_argument('nbins', type=int, help='number of bins')
    parser.add_argument('-i0', '--include_zeros', help='Include zeros in the analysis (always true for CIFTI)')
    grp = parser.add_mutually_exclusive_group()
    grp.add_argument('--equal_weight', type=cmd.greyordinate_in,
                     help='Each bin will contain the same total weight of given file')
    grp.add_argument('--equal_number', action='store_true',
                     help='Each bin will contain the same number of elements')
    grp.add_argument('--equal_bin', action='store_true',
                     help='Each bin will have the same size (from min to max value)')
    grp.add_argument('--set_bin', nargs='*', type=float,
                     help='Manually sets the edges of the bins as a sequence of numbers')



"""Converts between CIFTI formats (NIFTI, HDF5, and zarr)"""
import dask.array as da
import numpy as np
import zarr
from loguru import logger

from mcot.cifti.greyordinate import GreyOrdinates


def copy(source: GreyOrdinates, target: GreyOrdinates):
    """Copies information from source to target."""
    if source.data.shape != target.data.shape:
        raise ValueError("Source and target shape do not match")

    if isinstance(target.data, zarr.Array):
        target.data[:] = source.data
    else:
        chunks = getattr(
            target.data, 'chunks', getattr(
                source.data, 'chunks', 'auto'
            )
        )
        for dataset in target.data, source.data:
            if not hasattr(dataset, 'chunks'):
                if chunks == 'auto':
                    chunks = [1, 1]
                else:
                    chunks = list(chunks)
                chunks[np.argmin(dataset.strides)] = None
        logger.info(f"Adopted chunk size: {tuple(chunks)} for CIFTI with shape {tuple(source.data.shape)}")

        data = source.as_dask(tuple(chunks))
        da.store(data, target.data)


def run_from_args(args):
    """Runs script from command line arguments.

    :param args: command line arguments from :func:`add_to_parser`.
    """
    go_source = GreyOrdinates.from_filename(args.input)
    dtype = go_source.data.dtype if args.dtype is None else args.dtype
    logger.info("Creating empty output file")
    with GreyOrdinates.empty(
            args.output, go_source.other_axes + (go_source.brain_model_axis, ), dtype=dtype
    ) as go_target:
        copy(go_source, go_target)


def add_to_parser(parser):
    """Creates the parser of the command line arguments.

    After parsing the script can be run using :func:`run_from_args`.

    :param parser: parser to add arguments to
    """
    parser.add_argument('input', help='source NIFTI, GIFTI, CIFTI, HDF5, or zarr file with CIFTI data')
    parser.add_argument('output', help='target CIFTI, HDF5, or zarr filename')
    parser.add_argument('-t', '--dtype', help='sets the output datatype (default: keep same as input)')



#!/usr/bin/env python
"""Clusters the surface based on provided features."""
import numpy as np
from loguru import logger
from sklearn import cluster as skcluster

from mcot.cifti._scripts.cifti import gradient
from mcot.cifti.greyordinate import GreyOrdinates
from mcot.surface import CorticalMesh, write_gifti


def cluster(similarity, method='spectral', n_clusters=10, **kwargs):
    """Cluster based on a similarity graph.

    :param similarity: (N, N) sparse similarity matrix
    :param method: clustering method ('spectral', 'affinity', or 'DBSCAN')
    :param n_clusters: number of clusters
    :return: (N, ) integer array of labels
    """
    logger.info("Starting to cluster %s array using %s clustering", similarity.shape, method)
    if method.lower() == 'spectral':
        algorithm = skcluster.SpectralClustering(n_clusters=n_clusters, affinity='precomputed', **kwargs)
    elif method.lower() == 'affinity':
        algorithm = skcluster.AffinityPropagation(affinity='precomputed', preference=np.mean(similarity), **kwargs)
    elif method.lower() == 'dbscan':
        algorithm = skcluster.DBSCAN(metric='precomputed', n_clusters=n_clusters, **kwargs)
    return algorithm.fit_predict(similarity)


def run(surface, features, metric='spearman', method='spectral', n_clusters=10):
    """Clusters the surface based on the provided features.

    :param surface: anatomical surface with N vertices
    :param features: (M, N) array of surface features
    :param metric: which metric to use for the similarity between neighbouring vertices
    :param method: clustering method ('spectral', 'affinity', or 'DBSCAN')
    :param n_clusters: number of clusters
    :return: (N, ) integer array of labels
    """
    sim_graph = gradient.distance(surface.graph_point_point(include_diagonal=False), features, metric=metric)
    return cluster(sim_graph.tocsr(), method=method, n_clusters=n_clusters)


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    surface = CorticalMesh.read(args.surface)
    go = GreyOrdinates.from_filename(args.features)
    features = go.surface(surface.anatomy)
    mask = np.isfinite(features)
    labels = run(
            surface[mask],
            features[..., mask],
            metric=args.metric,
            method=args.cluster,
            n_clusters=args.n_clusters,
    )
    full_labels = -np.ones(surface.nvertices, dtype=labels.dtype)
    full_labels[mask] = labels
    write_gifti(args.output, [full_labels],
                brain_structure=surface.anatomy,
                color_map={-1: ('???', (0, 0, 0, 0))})


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument("surface", help='.surf.gii file describing the surface (WM/GM boundary, mid, or pial)')
    parser.add_argument("features", help='GIFTI/CIFTI file with the features of interest')
    parser.add_argument("output", help='.label.gii GIFTI file with the cluster labels')
    parser.add_argument("-m", '--metric', choices=gradient.dist_options, default='euclidean',
                        help='how to compute the similarity with the neighbours (default: euclidean)')
    parser.add_argument("-c", '--cluster', choices=('spectral', 'affinity', 'dbscan'), default='spectral',
                        help='Clustering algorithm to use')
    parser.add_argument("-n", "--n_clusters", type=int, default=10,
                        help='number of clusters to fit (ignored by affinity)')



#!/usr/bin/env python
"""Converts probtrackX to dconn."""
import os.path as op

import nibabel as nib
import numpy as np
import pandas as pd
from loguru import logger
from nibabel import cifti2, gifti
from nibabel.filebasedimages import ImageFileError
from scipy import sparse

from mcot.cifti.greyordinate import GreyOrdinates
from mcot.surface.cortical_mesh import CorticalMesh, get_brain_structure


def dot2matrix(dot_file):
    """Converts a fdt_matrix3.dot file into a sparse matrix.

    :param dot_file: dot-file
    :return: (N, N) matrix
    """
    logger.debug(f'loading dot-matrix from {dot_file}')
    indices = pd.read_csv(dot_file, delim_whitespace=True, dtype=int).values.T
    shape = indices[:-1, -1]
    indices = indices[:, :-1]
    return sparse.coo_matrix((indices[2], (indices[0], indices[1])), shape=shape)


def run(output_file, dot_file, brain_model, other_brain_model=None):
    """Creates a greyordinate object with the matrix data stored on disk.

    :param output_file: output filename
    :param dot_file: dot-file containing the matrix
    :param brain_model: Defines greyordinates along second axis (i.e., first column in dot file)
    :param other_brain_model: Defines greyordinates along first axis
        (i.e., second column in dot file; defaults to brain_model)
    """

    with GreyOrdinates.empty(output_file, (other_brain_model, brain_model), dtype=int) as go:
        nrows = min(len(brain_model), int(2e8) // (len(other_brain_model) * 4))
        if hasattr(go.data, 'chunks') and go.data.chunks[1] < nrows:
            nrows = (nrows // go.data.chunks[1]) * go.data.chunks[1]
        logger.info(f"Storing {nrows} rows into memory before writing to disk")
        tmp_arr = np.zeros((nrows, len(other_brain_model)), dtype=int)

        current_idx = 0

        for idx, df in enumerate(pd.read_csv(dot_file, delim_whitespace=True, dtype=int,
                                             chunksize=int(1e6), header=None)):
            row, col, data = df.values.T
            if data[-1] == 0:
                row, col, data = row[:-1], col[:-1], data[:-1]
                print(row[-1], col[-1], data[-1])
                logger.info("Reached final row")
            assert (row[:-1] <= row[1:]).all()
            upper = np.searchsorted(row, current_idx + 1, 'right')
            tmp_arr[current_idx % nrows, col[:upper] - 1] = data[:upper]
            assert upper == 0 or row[0] == row[upper - 1]
            while row[-1] > current_idx + 1:
                current_idx += 1
                if current_idx % nrows == 0:
                    logger.debug(f'storing rows up to {current_idx}')
                    go.data[:, current_idx - nrows: current_idx] = tmp_arr.T
                    tmp_arr[()] = 0

                lower = upper
                upper = np.searchsorted(row[lower:], current_idx + 1, 'right') + lower
                assert (lower == upper) or (row[lower] == row[upper - 1])
                tmp_arr[current_idx % nrows, col[lower:upper] - 1] = data[lower:upper]

        nused = (current_idx % nrows) + 1
        go.data[:, current_idx - nused + 1:current_idx + 1] = tmp_arr[:nused].T

        logger.debug(f"Final row stored from {current_idx - nused + 1} till {current_idx + 1}")


def get_brain_model(filename: str, assumed_hemis=None) -> cifti2.BrainModelAxis:
    """Creates a CIFTI BrainModel axis based on the provided file.

    :param filename: can be one of the following:

        - dense CIFTI file, whose BrainModel axis should be adopted
        - single volume (NIFTI) or surface (GIFTI/ASCII) used as a mask
        - list of volumes and surfaces

    :param assumed_hemis: editable list of hemispheres each surface is assumed to be
    :return: BrainModelAxis describing the dataset
    """
    try:
        if not isinstance(filename, str):
            img = filename
        else:
            img = nib.load(filename)
    except ImageFileError:
        with open(filename, 'r') as f:
            first_line = f.read()
        if first_line.startswith('#!ascii from CsvMesh'):
            return read_ascii(filename, None if assumed_hemis is None else assumed_hemis.pop(0))
        with open(filename, 'r') as f:
            bm_parts = []
            for line in f.readlines():
                single_filename = line.strip()
                if len(single_filename) > 0:
                    if not op.isfile(single_filename):
                        raise IOError(f"Mask filename {single_filename} not found")
                    bm_parts.append(get_brain_model(single_filename))
        if len(bm_parts) == 0:
            raise ValueError(f"No masks found in {filename}")
        bm = bm_parts[0]
        for part in bm_parts[1:]:
            bm = bm + part
        return bm

    # filename was loaded as a nibabel image
    if isinstance(img, cifti2.Cifti2Image):
        return img.header.get_axis(img.ndim - 1)
    elif isinstance(img, gifti.GiftiImage):
        mask = img.darrays[-1].data
        if mask.ndim == 2:
            mask = np.ones(CorticalMesh.read(img).nvertices, dtype='bool')
        return cifti2.BrainModelAxis.from_mask(mask, name=get_brain_structure(img).cifti)
    else:
        transposed_mask = np.transpose(img.get_fdata() > 0, (2, 1, 0))
        bm = cifti2.BrainModelAxis.from_mask(transposed_mask, affine=img.affine)
        bm.voxel = bm.voxel[:, ::-1]
        return bm


def read_ascii(filename, assumed_hemi=None):
    """Reads a surface stores as an ASCII file.

    :param filename: ASCII file to read
    :param assumed_hemi: L or R defining the assumed hemisphere (default: guess from filename)
    :return: list of filenames in the ASCII file (or ASCII filename itself)
    """
    if assumed_hemi not in (None, 'L', 'R'):
        raise ValueError(f"Assumed hemisphere should be set to 'L' or 'R', not {assumed_hemi}")

    with open(filename, 'r') as f:
        first_line = f.readline()
        assert first_line.startswith('#!ascii from CsvMesh')
        nvertices, nfaces = [int(part) for part in f.readline().strip().split()]
        mask = np.zeros(nvertices, dtype='int')
        for idx in range(nvertices):
            mask[idx] = int(f.readline().strip().split()[-1])

    if assumed_hemi is not None:
        structure = 'CortexLeft' if assumed_hemi == 'L' else 'CortexRight'
    else:
        final_part = op.split(filename)[-1]
        if 'L' in final_part and 'R' in final_part:
            raise ValueError("Does not know which hemisphere %s belongs to" % filename)
        if 'L' in final_part:
            structure = 'CortexLeft'
        elif 'R' in final_part:
            structure = 'CortexRight'
        else:
            raise ValueError("Does not know which hemisphere %s belongs to" % filename)
    logger.info(f'{filename} interpreted as {structure}')

    return cifti2.BrainModelAxis.from_mask(mask, name=structure)


def run_from_args(args):
    logger.info('starting %s', op.basename(__file__))
    assumed_hemis = None if args.hemi is None else list(args.hemi)
    brain_model = get_brain_model(args.targets, assumed_hemis)
    other_brain_model = brain_model if args.otargets is None else get_brain_model(args.otargets, assumed_hemis)

    run(
        output_file=args.output,
        dot_file=args.dot_file,
        brain_model=brain_model,
        other_brain_model=other_brain_model
    )
    logger.info('ending %s', op.basename(__file__))


def add_to_parser(parser):
    parser.add_argument('dot_file', help='fdt_matrix.dot output from probtrackx')
    parser.add_argument('targets', help='CIFTI with brain model axis to use or '
                                        'ASCII file with target images (or single volume/surface file). '
                                        'Seed mask in case of matrix2')
    parser.add_argument('output', help='output .dconn.nii dense connectome file')
    parser.add_argument('--otargets', help='Target of matrix2 or --otargets for matrix3 (defaults to targets)')
    parser.add_argument('--hemi', nargs='+',
                        help='hemispheres to assume for the surface ASCII files (sequence of L or R). '
                             'First define for `targets` than `otargets`.')



#!/usr/bin/env python
"""Randomly parcellates a surface."""
import os.path as op

import nibabel as nib
import numpy as np
from loguru import logger
from scipy.sparse import csgraph

from mcot.surface import CorticalMesh, write_gifti


def init_random(distance):
    """Random initialization of the medoid centers."""
    indices = np.arange(distance.shape[0], dtype='int')
    np.random.shuffle(indices)
    return indices


def init_center(distance):
    """Initalize medoid centers with the points with the lowest total
    distance."""
    dist_normed = distance / np.sum(distance, -1)[:, None]
    total_dist_normed = np.sum(dist_normed, 0)
    return np.argsort(total_dist_normed)


def update_centers(distance, labels):
    """Updates the cluster centers."""
    nclus = np.unique(labels).size
    centers = np.zeros(nclus, dtype='int')
    for label in range(nclus):
        use = label == labels
        subdist = distance[use, :][:, use]
        centers[label] = np.where(use)[0][np.argmin(np.sum(subdist, -1))]
    return centers


def update_labels(distance, centers):
    """Update the point labels."""
    return np.argmin(distance[:, centers], -1)


def calc_distance(distance, centers, labels):
    """Calculate the distance between the cluster centra and the labels."""
    return sum([sum(distance[centers[label], labels == label]) for label in range(centers.size)])


def kmedoids(distance, nclus=10, init='center', max_loops=6):
    """Runs the k-medoids clustering algorithm.

    based on Park & Jun (2008): http://www.sciencedirect.com/science/article/pii/S095741740800081X

    :param distance: NxN distance matrix
    :param nclus: number of clusters
    :param init: initialization ('random' or 'center')
    :param max_loops: maximum number of loops
    :return: tuple with

        - `centers`: indices of the cluster centers
        - `labels`: N-length integer array of cluster labels
    """
    if init == 'random':
        centers = init_random(distance)[:nclus]
    elif init == 'center':
        centers = init_center(distance)[:nclus]
    else:
        centers = init
    old_labels = -np.ones(distance.shape[0])
    for idx_loop in range(max_loops):
        print(idx_loop)
        print('updating')
        labels = update_labels(distance, centers)
        if (labels == old_labels).all():
            logger.info('Found optimum after {} loops'.format(idx_loop + 1))
            break
        old_labels = labels
        print('update centres')
        centers = update_centers(distance, labels)
    else:
        logger.warning("Failed to find maximum")
    return centers, labels


def uniform_centroids(dist_map, n_centroids):
    """Uniformly space `n_centroids` seeds in a naive way.

    :param dist_map: sparse distance map
    :param n_centroids: number of seeds to place
    :return: (n_centroids, ) integer arrays with the indices of the seeds
    """
    def get_dist(idx_vertex):
        return csgraph.dijkstra(dist_map, indices=idx_vertex, directed=False)

    res = np.zeros(n_centroids, dtype='i4')
    res[0] = np.random.randint(0, dist_map.shape[0])
    dist = get_dist(res[0])
    for idx in range(1, n_centroids):
        res[idx] = np.argmax(dist)
        np.minimum(dist, get_dist(res[idx]), out=dist)
    return res


def run(surface: CorticalMesh, ncluster: int, max_iter=0):
    """Creates a random parcellation of the surface based on the distance
    between surface elements.

    :param surface: cortical surface with N vertices
    :param ncluster: number of clusters
    :param max_iter: maximum number of iterations in the k-medoids (if 0 a fast voronoi evaluation is used)
    :return: (N, ) integer array with values from 0 to N-1
    """
    sparse_dist_map = surface.graph_point_point(weight='distance', dtype='f8')
    if max_iter == 0:
        centroids = uniform_centroids(sparse_dist_map, ncluster)
        dens_dist_map = csgraph.dijkstra(sparse_dist_map, indices=centroids, directed=False)
        return dens_dist_map.argmin(0)
    else:
        dense_dist_map = csgraph.dijkstra(sparse_dist_map, directed=False)
        return kmedoids.kmedoids(dense_dist_map, ncluster, max_loops=max_iter)[1]


def run_from_args(args):
    """Runs the script based on a Namespace containing the command line
    arguments."""
    logger.info('starting %s', op.basename(__file__))
    mask = nib.load(args.mask).darrays[0].data > 0
    full_surface = CorticalMesh.read(args.surface)
    labels = run(
            surface=full_surface[mask],
            ncluster=args.ncluster,
            max_iter=args.iter,
    )
    full_labels = np.zeros(full_surface.nvertices, dtype='i4')
    full_labels[mask] = labels + 1

    write_gifti.write_gifti(args.out, [full_labels], brain_structure=full_surface.anatomy,
                            color_map={0: ('unk', (0, 0, 0, 0))})
    logger.info('ending %s', op.basename(__file__))


def add_to_parser(parser):
    """Creates the parser of the command line arguments."""
    parser.add_argument("surface", help='.surf.gii file with the surface')
    parser.add_argument("mask", help='.shape.gii file with the surface mask')
    parser.add_argument("ncluster", type=int, help='number of clusters')
    parser.add_argument("out", help='.dlabel.gii dense GIFTI label output (0 outside of mask)')
    parser.add_argument("--iter", default=0, type=int, help='maximum number of iterations (default: 0)')



"""Common interface to run all of the scripts."""
import os.path as op

from mcot.utils.scripts import ScriptDirectory


def main():
    """Runs one of the scripts from the command line."""
    ScriptDirectory(op.split(__file__)[0])()


if __name__ == '__main__':
    main()